#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['Dimensionless', 'Percentage']

from unyts.dictionaries import dictionary
from unyts.unit_class import Unit
from unyts.errors import WrongUnitsError
from unyts.helpers.common_classes import unit_or_str, numeric


class Dimensionless(Unit):
    classUnits = dictionary['Dimensionless']

    def __init__(self, value: numeric, units: unit_or_str = None) -> Unit:
        self.name = 'dimensionless'
        self.kind = Dimensionless
        self.value = self.check_value(value)
        if units is None:
            units = 'dimensionless'
        self.unit = self.check_unit(units)

    def convert(self, new_unit: unit_or_str = None) -> Unit:
        if new_unit is not None and type(new_unit) is not str:
            try:
                new_unit = new_unit.unit
            except:
                raise WrongUnitsError("'" + str(new_unit) + "' for '" + str(self.name) + "'")
        if new_unit is None or len(new_unit) == 0:
            return self.value
        elif new_unit in dictionary['Percentage']:
            return Percentage(self.value * 100, new_unit)
        elif new_unit in dictionary['Dimensionless']:
            return Dimensionless(self.value, new_unit)
        else:
            from .define import units
            return units(self.value, new_unit)

    def to(self, new_unit: unit_or_str = None) -> Unit:
        return self.convert(new_unit)


class Percentage(Dimensionless):
    classUnits = dictionary['Percentage']

    def __init__(self, value: numeric, units: unit_or_str = None) -> Unit:
        self.name = 'percentage'
        self.kind = Percentage
        self.value = self.check_value(value) / 100
        if units is None:
            units = 'percent'
        self.unit = self.check_unit(units)

    def __repr__(self) -> str:
        return str(self.value * 100) + '_' + str(self.unit)

    def __str__(self) -> str:
        return str(self.value * 100) + '_' + str(self.unit)

    def __neg__(self) -> Unit:
        return self.kind(self.value * -100, self.unit)

    def __abs__(self) -> Unit:
        return self.kind(abs(self.value) * 100, self.unit)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['array_like', 'number', 'numeric', 'unit_or_str']

from numpy import ndarray
from typing import Union
from unyts.unit_class import Unit

try:
    from pandas import Series, DataFrame
    array_like = tuple([ndarray, Series, DataFrame])
    numeric = Union[int, float, complex, ndarray, Series, DataFrame]
except ModuleNotFoundError:
    array_like = tuple([ndarray])
    numeric = Union[int, float, complex, ndarray]
number = Union[int, float, complex]
unit_or_str = Union[Unit, str]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:38:58 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['unit_product', 'unit_division']

from unyts.dictionaries import dictionary
from unyts.converter import convertible
from unyts.helpers.is_number import is_number
from unyts.helpers.multi_split import multi_split


def unit_split(unit_string: str) -> str:
    us = multi_split(unit_string)
    return us


def unit_base_power(unit_string: str) -> tuple[str, int]:
    u_bas, u_pow = '', ''
    oth = ''
    inv = False
    if '/' in unit_string and len(unit_string.split('/')) == 2 and is_number(unit_string.split('/')[0]):
        inv_pow, unit_string = unit_string.split('/')
        inv = True

    for c in unit_string:
        if c.isdigit():
            u_pow += oth + c
            oth = ''
        elif c in ['-', '+', '.']:
            oth += c
        else:
            u_bas += oth + c
            oth = ''
    u_pow = 1 if u_pow == '' else float(u_pow) if '.' in u_pow else int(u_pow)
    if inv:
        u_pow = u_pow * -1 * (float(inv_pow) if '.' in inv_pow else int(inv_pow))
    return u_bas, u_pow


def unit_base(unit_string: str) -> str:
    return unit_base_power(unit_string)[0]


def unit_product(unit_string1: str, unit_string2: str) -> str:
    if unit_string1 is None:
        unit_string1 = 'dimensionless'
    if unit_string2 is None:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.strip(' ()')) == 0:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.strip(' ()')) == 0:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.split('/')) == 2 and unit_string1.split('/')[0] == \
            unit_string1.split('/')[1]:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.split('/')) == 2 and unit_string2.split('/')[0] == \
            unit_string2.split('/')[1]:
        unit_string2 = 'dimensionless'

    if unit_string2.lower().strip(' ()') in dictionary['Dimensionless']:
        return unit_string1
    if unit_string1.lower().strip(' ()') in dictionary['Dimensionless']:
        if unit_string2.lower().strip(' ()') not in dictionary['Dimensionless']:
            return unit_string2
        else:
            return unit_string1

    if unit_string1 != unit_string2 and convertible(unit_string1, unit_string2):
        return unit_product(unit_string1, unit_string1)

    u1bas, u1pow = unit_base_power(unit_string1)
    u2bas, u2pow = unit_base_power(unit_string2)

    if convertible(u1bas, u2bas):
        u_pow = u1pow + u2pow
        if u_pow == -1:
            result = u1bas + '-1'
        elif u_pow == 1:
            result = u1bas
        elif u_pow == 0:
            result = u1bas + '/' + u1bas
        else:
            for c in ['+', '-', '^']:  # '*','/'
                if c in u1bas:
                    u1bas = '(' + u1bas + ')'
                    break
            result = u1bas + str(u_pow)
    else:
        for c in ['+', '-', '^']:  # '*','/'
            if c in unit_string1:
                unit_string1 = '(' + unit_string1 + ')'
                break
        for c in ['+', '-', '^']:  # '*','/'
            if c in unit_string2:
                unit_string2 = '(' + unit_string2 + ')'
                break
        result = unit_string1 + '*' + unit_string2
    return result


def unit_division(unit_string1: str, unit_string2: str) -> str:
    if unit_string1 is None:
        unit_string1 = 'dimensionless'
    if unit_string2 is None:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.strip(' ()')) == 0:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.strip(' ()')) == 0:
        unit_string2 = 'dimensionless'

    if type(unit_string1) is str and len(unit_string1.split('/')) == 2 and unit_string1.split('/')[0] == \
            unit_string1.split('/')[1]:
        unit_string1 = 'dimensionless'
    if type(unit_string2) is str and len(unit_string2.split('/')) == 2 and unit_string2.split('/')[0] == \
            unit_string2.split('/')[1]:
        unit_string2 = 'dimensionless'

    if unit_string2.lower().strip(' ()') in dictionary['Dimensionless']:
        return unit_string1
    if unit_string1.lower().strip(' ()') in dictionary['Dimensionless']:
        if unit_string2.lower().strip(' ()') not in dictionary['Dimensionless']:
            u_bas, u_pow = unit_base_power('1/' + unit_string2)
            return u_bas + str(u_pow)
        else:
            return unit_string1

    if unit_string1 != unit_string2 and convertible(unit_string1, unit_string2):
        return unit_division(unit_string1, unit_string1)

    u1bas, u1pow = unit_base_power(unit_string1)
    u2bas, u2pow = unit_base_power(unit_string2)

    if convertible(u1bas, u2bas):
        u_pow = u1pow - u2pow
        if u_pow == -1:
            result = u1bas + '-1'
        elif u_pow == 1:
            result = u1bas
        elif u_pow == 0:
            result = u1bas + '/' + u1bas
        else:
            for c in ['+', '-', '^']:  # '*','/'
                if c in u1bas:
                    u1bas = '(' + u1bas + ')'
                    break
            result = u1bas + str(u_pow)

    elif ('+' not in unit_string1 and '-' not in unit_string1 and '^' not in unit_string1) and (
            '+' not in unit_string2 and '-' not in unit_string2 and '^' not in unit_string2) and (
            '*' in unit_string1 and unit_base(unit_string2) in map(unit_base, unit_string1.split('*'))):
        result = ''
        for u in unit_string1.split('*'):
            if unit_string2 == u:
                return '*'.join([u for u in unit_string1.split('*') if u != unit_string2])
            elif unit_base(unit_string2) == unit_base(u):
                result = (result + '*' + unit_division(u, unit_string2).strip('*')).strip('*')
            else:
                result = (result + '*' + u).strip('*')
    else:
        for c in ['+', '-', '^']:  # '*','/'
            if c in unit_string1:
                unit_string1 = '(' + unit_string1 + ')'
                break
        for c in ['+', '-', '^']:  # '*','/'
            if c in unit_string2:
                unit_string2 = '(' + unit_string2 + ')'
                break
        result = unit_string1 + '/' + unit_string2
    return result


def unit_power(unit_string: str, power: int or str) -> str:
    if unit_string is None:
        unit_string = 'dimensionless'
    if power is None:
        power = 1

    if type(unit_string) is str and len(unit_string.strip(' ()')) == 0:
        unit_string = 'dimensionless'
    if type(power) is str and len(power.strip(' ()')) == 0:
        power = 1

    u1bas, u1pow = unit_base_power(unit_string)

    if type(power) in (int, float):
        u_pow = u1pow * power
        if u_pow == 0:
            return 'dimensionless'
        elif u_pow == 1:
            return u1bas
        else:
            return u1bas + str(u_pow)
    elif type(power) is str:
        u2bas, u2pow = unit_base_power(power)
        u_pow = u1pow * u2pow
        if u_pow == 0:
            return 'dimensionless'
        elif u_pow == 1:
            return u1bas + '^' + u2bas
        else:
            return u1bas + '^' + str(u_pow) + u2bas
    else:
        raise TypeError('Power must be units string or numeric.')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['Length', 'Area', 'Volume']

from unyts.dictionaries import dictionary
from unyts.unit_class import Unit
from unyts.helpers.common_classes import unit_or_str, numeric


class Length(Unit):
    classUnits = dictionary['Length']

    def __init__(self, value: numeric, units: unit_or_str) -> Unit:
        self.name = 'length'
        self.kind = Length
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class Area(Unit):
    classUnits = dictionary['Area']

    def __init__(self, value: numeric, units: unit_or_str) -> Unit:
        self.name = 'area'
        self.kind = Area
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class Volume(Unit):
    classUnits = dictionary['Volume']

    def __init__(self, value: numeric, units: unit_or_str) -> Unit:
        self.name = 'volume'
        self.kind = Volume
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['Density', 'VolumeRatio', 'ProductivityIndex', 'PressureGradient']

from unyts.dictionaries import dictionary
from unyts.unit_class import Unit
from unyts.helpers.common_classes import unit_or_str, numeric


class Density(Unit):
    classUnits = dictionary['Density']

    def __init__(self, value: numeric, units: unit_or_str) -> Unit:
        self.name = 'density'
        self.kind = Density
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class VolumeRatio(Unit):
    classUnits = dictionary['VolumeRatio']

    def __init__(self, value: numeric, units: unit_or_str) -> Unit:
        self.name = 'volumeRatio'
        self.kid = VolumeRatio
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class ProductivityIndex(Unit):
    classUnits = dictionary['ProductivityIndex']

    def __init__(self, value: numeric, units: unit_or_str) -> Unit:
        self.name = 'productivityIndex'
        self.kind = ProductivityIndex
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)


class PressureGradient(Unit):
    classUnits = dictionary['PressureGradient']

    def __init__(self, value: numeric, units: unit_or_str) -> Unit:
        self.name = 'pressureGradient'
        self.kind = PressureGradient
        self.value = self.check_value(value)
        self.unit = self.check_unit(units)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 17:52:34 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['BFS', 'print_path']


def BFS(graph, start, end, verbose=False) -> list:
    """Assumes graph is a Digraph; start and end are nodes
        Returns a shortest path from start to end in graph"""
    init_path = [start]
    path_queue = [init_path]
    visited = []
    while len(path_queue) != 0:
        # get and remove oldest element in path_queue
        conv_path = path_queue.pop(0)
        if conv_path in visited:
            if verbose:
                print('<UnitsConv> ' + str(len(path_queue)) + ' paths in queue. ' + 'Already visited BFS path:\n', print_path(conv_path))
        else:
            if verbose:
                print('<UnitsConv> ' + str(len(path_queue)) + ' paths in queue. ' + 'Current BFS path:\n', print_path(conv_path))
            last_node = conv_path[-1]
            if last_node is end:
                return conv_path
            for next_node in graph.children_of(last_node):
                if next_node not in conv_path:
                    new_path = conv_path + [next_node]
                    path_queue.append(new_path)
            visited.append(conv_path)


def print_path(path: list) -> str:
    """Assumes path is a list of nodes"""
    result = '    '
    if len(path) == 1:
        result = result + str(path[0]) + ' = ' + str(path[0])
    else:
        for i in range(len(path)):
            if type(path[i]) is str and path[i] not in ['(1)', '(v)']:
                if result[-3:] == ' > ':
                    result = result[:-3]
                result = result + ' ' + path[i] + ' '
            elif type(path[i]) in (int, float, complex):
                result = result + str(path[i]) + ' '
            else:
                result = result + str(path[i])
                if i != len(path) - 1:
                    result = result + ' > '
    return result.strip()



# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:46:04 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['is_unit']

from unyts.dictionaries import dictionary


def is_unit(unit: str) -> bool:
    if type(unit) is str:
        unit = unit.strip()
    else:
        raise TypeError("unit must be a string.")

    for each in list(dictionary.keys()):
        if '_' not in each:
            is_u = unit in list(dictionary[each])
            if is_u:
                return True

    if '/' in unit or '*' in unit:
        unit_split = []
        for each in unit.split('/'):
            unit_split += each.split('*')
        ret = [False] * len(unit_split)
        for each in list(dictionary.keys()):
            if '_' not in each:
                for sub_u in range(len(unit_split)):
                    if unit_split[sub_u] in list(dictionary[each]):
                        ret[sub_u] = True
        for each in ret:
            if not each:
                return False
        return True
    return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 26 21:17:35 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['Energy', 'Power']

from unyts.dictionaries import dictionary
from unyts.unit_class import Unit
from unyts.helpers.common_classes import unit_or_str, numeric


class Energy(Unit):
    classUnits = dictionary['Energy']

    def __init__(self, value: numeric, unit: unit_or_str) -> Unit:
        self.name = 'energy'
        self.kind = Energy
        self.value = self.check_value(value)
        self.unit = self.check_unit(unit)


class Power(Unit):
    classUnits = dictionary['Power']

    def __init__(self, value: numeric, unit: unit_or_str) -> Unit:
        self.name = 'power'
        self.kind = Power
        self.value = self.check_value(value)
        self.unit = self.check_unit(unit)



# -*- coding: utf-8 -*-
"""
Created on Sun May 23 11:18:15 2021

@author: Martín Carlos Araya <martinaraya@gmail.com>

helper functions for units modules
"""

__version__ = '0.4.7'
__release__ = 20221229
__all__ = ['is_number']


def is_number(string: str) -> bool:
    """
    Checks if a string represents a number.
    Valid numbers are:
        positive integers, as in .isnumeric() and .isdigit() methods from string class
        floats, numbers with pointsç
        negative numbers, with dash at the left of the digits
        complex numbers,

    Parameters
    ----------
    string : str
        the string to check if represents a numbers

    Returns
    -------
    True or False

    """
    try:
        complex(string)
        return True
    except ValueError:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 12:36:48 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['unitsNetwork']

from unyts.dictionaries import SI, SI_order, OGF, OGF_order, DATA, DATA_order, dictionary, StandardAirDensity, \
    StandardEarthGravity
from unyts.network import UDigraph, UNode, Conversion
from unyts.parameters import unyts_parameters_, dir_path
from cloudpickle import dump, load
from os.path import isfile
from json import dump as jdump
from pandas import DataFrame


def _load_network():
    print('preparing units network...')
    network = UDigraph()

    for unit_kind in list(dictionary.keys()):
        # print('1: ' +unit_kind)
        if '_' not in unit_kind:
            for unit_name in dictionary[unit_kind]:
                # print('_ 2: ' + unit_name)
                network.add_node(UNode(unit_name))
        if '_NAMES' in unit_kind:
            for unit_name in list(dictionary[unit_kind].keys()):
                # print('N  2: ' + unit_name,unit_kind.split('_')[0])
                network.add_node(UNode(unit_name))
                dictionary[unit_kind.split('_')[0]].append(unit_name)
                for secondName in dictionary[unit_kind][unit_name]:
                    # print('N   3: ' + unit_name)
                    network.add_node(UNode(secondName))
                    network.add_edge(Conversion(network.get_node(secondName), network.get_node(unit_name), lambda x: x))
                    network.add_edge(Conversion(network.get_node(unit_name), network.get_node(secondName), lambda x: x))
                    dictionary[unit_kind.split('_')[0]].append(secondName)
        if '_SPACES' in unit_kind:
            for unit_name in list(dictionary[unit_kind].keys()):
                # print('N  2: ' + unit_name,unit_kind.split('_')[0])
                if ' ' in unit_name:
                    network.add_node(UNode(unit_name))
                    network.add_node(UNode(unit_name.replace(' ', '-')))
                    dictionary[unit_kind.split('_')[0]].append(unit_name)
                    dictionary[unit_kind.split('_')[0]].append(unit_name.replace(' ', '-'))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(unit_name.replace(' ', '-')),
                                   lambda x: x))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(unit_name.replace(' ', '-')),
                                   lambda x: x))
                    for secondName in dictionary[unit_kind][unit_name]:
                        # print('N   3: ' + unit_name)
                        if ' ' in secondName:
                            network.add_node(UNode(secondName))
                            network.add_node(UNode(secondName.replace(' ', '-')))
                            network.add_edge(
                                Conversion(network.get_node(secondName.replace(' ', '-')), network.get_node(secondName),
                                           lambda x: x))
                            network.add_edge(
                                Conversion(network.get_node(secondName), network.get_node(secondName.replace(' ', '-')),
                                           lambda x: x))
                            dictionary[unit_kind.split('_')[0]].append(secondName)
                            dictionary[unit_kind.split('_')[0]].append(secondName.replace(' ', '-'))
                else:
                    for secondName in dictionary[unit_kind][unit_name]:
                        # print('N   3: ' + unit_name)
                        if ' ' in secondName:
                            network.add_node(UNode(secondName))
                            network.add_node(UNode(secondName.replace(' ', '-')))
                            network.add_edge(
                                Conversion(network.get_node(secondName.replace(' ', '-')), network.get_node(secondName),
                                           lambda x: x))
                            network.add_edge(
                                Conversion(network.get_node(secondName), network.get_node(secondName.replace(' ', '-')),
                                           lambda x: x))
                            dictionary[unit_kind.split('_')[0]].append(secondName)
                            dictionary[unit_kind.split('_')[0]].append(secondName.replace(' ', '-'))

        if '_SI' in unit_kind and unit_kind.split('_')[0] in SI_order[0]:
            for unit_name in list(dictionary[unit_kind]):
                # print('S  2: ' + unit_name)
                network.add_node(UNode(unit_name))
                dictionary[unit_kind.split('_')[0]].append(unit_name)
                for prefix in list(SI.keys()):
                    # print('S   3: ' + prefix+unit_name+'_'+str(SI[prefix][0]))
                    network.add_node(UNode(prefix + unit_name))
                    network.add_edge(
                        Conversion(network.get_node(prefix + unit_name), network.get_node(unit_name), SI[prefix][0]))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(prefix + unit_name), SI[prefix][0],
                                   True))
                    dictionary[unit_kind.split('_')[0]].append(prefix + unit_name)
        if '_SI' in unit_kind and unit_kind.split('_')[0] in SI_order[1]:
            for unit_name in list(dictionary[unit_kind]):
                # print('S  2: ' + unit_name)
                network.add_node(UNode(unit_name))
                dictionary[unit_kind.split('_')[0]].append(unit_name)
                for prefix in list(SI.keys()):
                    # print('S   3: ' + prefix+unit_name+'_'+str(SI[prefix][1]))
                    network.add_node(UNode(prefix + unit_name))
                    network.add_edge(
                        Conversion(network.get_node(prefix + unit_name), network.get_node(unit_name), SI[prefix][1]))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(prefix + unit_name), SI[prefix][1],
                                   True))
                    dictionary[unit_kind.split('_')[0]].append(prefix + unit_name)
        if '_SI' in unit_kind and unit_kind.split('_')[0] in SI_order[2]:
            for unit_name in list(dictionary[unit_kind]):
                # print('S  2: ' + unit_name)
                network.add_node(UNode(unit_name))
                dictionary[unit_kind.split('_')[0]].append(unit_name)
                for prefix in list(SI.keys()):
                    # print('S   3: ' + prefix+unit_name+'_'+str(SI[prefix]))
                    network.add_node(UNode(prefix + unit_name))
                    network.add_edge(
                        Conversion(network.get_node(prefix + unit_name), network.get_node(unit_name), SI[prefix][2]))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(prefix + unit_name), SI[prefix][2],
                                   True))
                    dictionary[unit_kind.split('_')[0]].append(prefix + unit_name)
        if '_DATA' in unit_kind and unit_kind.split('_')[0] in DATA_order[0]:
            for unit_name in list(dictionary[unit_kind]):
                # print('S  2: ' + unit_name)
                network.add_node(UNode(unit_name))
                dictionary[unit_kind.split('_')[0]].append(unit_name)
                for prefix in list(DATA.keys()):
                    # print('S   3: ' + prefix+unit_name+'_'+str(SI[prefix]))
                    network.add_node(UNode(prefix + unit_name))
                    network.add_edge(
                        Conversion(network.get_node(prefix + unit_name), network.get_node(unit_name), DATA[prefix][0]))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(prefix + unit_name), DATA[prefix][0],
                                   True))
                    dictionary[unit_kind.split('_')[0]].append(prefix + unit_name)
        if '_DATA' in unit_kind and unit_kind.split('_')[0] in DATA_order[1]:
            for unit_name in list(dictionary[unit_kind]):
                # print('S  2: ' + unit_name)
                network.add_node(UNode(unit_name))
                dictionary[unit_kind.split('_')[0]].append(unit_name)
                for prefix in list(DATA.keys()):
                    # print('S   3: ' + prefix+unit_name+'_'+str(SI[prefix]))
                    network.add_node(UNode(prefix + unit_name))
                    network.add_edge(
                        Conversion(network.get_node(prefix + unit_name), network.get_node(unit_name), DATA[prefix][1]))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(prefix + unit_name), DATA[prefix][1],
                                   True))
                    dictionary[unit_kind.split('_')[0]].append(prefix + unit_name)
        if '_OGF' in unit_kind and unit_kind.split('_')[0] in OGF_order[2]:
            for unit_name in list(dictionary[unit_kind]):
                # print('O  2: ' + unit_name)
                network.add_node(UNode(unit_name))
                dictionary[unit_kind.split('_')[0]].append(unit_name)
                for prefix in list(OGF.keys()):
                    # print('S   3: ' + prefix+unit_name+'_'+str(SI[prefix]))
                    network.add_node(UNode(prefix + unit_name))
                    network.add_edge(
                        Conversion(network.get_node(prefix + unit_name), network.get_node(unit_name), OGF[prefix][2]))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(prefix + unit_name), OGF[prefix][2],
                                   True))
                    dictionary[unit_kind.split('_')[0]].append(prefix + unit_name)
        if '_PLURALwS' in unit_kind:
            if type(dictionary[unit_kind]) is dict:
                list_names = list(dictionary[unit_kind].keys())
                for unit_name in list(dictionary[unit_kind].keys()):
                    # print('U  2: ' + unit_name,unit_kind.split('_')[0])
                    network.add_node(UNode(unit_name))
                    network.add_node(UNode(unit_name + 's'))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(unit_name + 's'), lambda x: x))
                    network.add_edge(
                        Conversion(network.get_node(unit_name + 's'), network.get_node(unit_name), lambda x: x))
                    dictionary[unit_kind.split('_')[0]].append(unit_name + 's')
            else:
                for unit_name in list(dictionary[unit_kind]):
                    # print('U  2: ' + unit_name,unit_kind.split('_')[0])
                    network.add_node(UNode(unit_name))
                    network.add_node(UNode(unit_name + 's'))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(unit_name + 's'), lambda x: x))
                    network.add_edge(
                        Conversion(network.get_node(unit_name + 's'), network.get_node(unit_name), lambda x: x))
                    dictionary[unit_kind.split('_')[0]].append(unit_name + 's')
            if '_UPPER' in unit_kind:
                if type(dictionary[unit_kind]) is dict:
                    list_names = list(dictionary[unit_kind].keys())
                    for unit_name in list(dictionary[unit_kind].keys()):
                        # print('U  2: ' + unit_name,unit_kind.split('_')[0])
                        network.add_node(UNode(unit_name))
                        network.add_node(UNode(unit_name.upper() + 'S'))
                        network.add_edge(
                            Conversion(network.get_node(unit_name), network.get_node(unit_name.upper() + 'S'),
                                       lambda x: x))
                        network.add_edge(
                            Conversion(network.get_node(unit_name.upper() + 'S'), network.get_node(unit_name),
                                       lambda x: x))
                        dictionary[unit_kind.split('_')[0]].append(unit_name.upper() + 'S')
                else:
                    for unit_name in list(dictionary[unit_kind]):
                        # print('U  2: ' + unit_name,unit_kind.split('_')[0])
                        network.add_node(UNode(unit_name))
                        network.add_node(UNode(unit_name.upper() + 'S'))
                        network.add_edge(
                            Conversion(network.get_node(unit_name), network.get_node(unit_name.upper() + 'S'),
                                       lambda x: x))
                        network.add_edge(
                            Conversion(network.get_node(unit_name.upper() + 'S'), network.get_node(unit_name),
                                       lambda x: x))
                        dictionary[unit_kind.split('_')[0]].append(unit_name.upper() + 'S')
        if '_UPPER' in unit_kind:
            if type(dictionary[unit_kind]) is dict:
                list_names = list(dictionary[unit_kind].keys())
                for unit_name in list(dictionary[unit_kind].keys()):
                    # print('U  2: ' + unit_name,unit_kind.split('_')[0])
                    network.add_node(UNode(unit_name))
                    network.add_node(UNode(unit_name.upper()))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(unit_name.upper()), lambda x: x))
                    network.add_edge(
                        Conversion(network.get_node(unit_name.upper()), network.get_node(unit_name), lambda x: x))
                    dictionary[unit_kind.split('_')[0]].append(unit_name.upper())
                    for secondName in dictionary[unit_kind][unit_name]:
                        # print('U   3: ' + unit_name)
                        network.add_node(UNode(secondName))
                        network.add_node(UNode(secondName.upper()))
                        network.add_edge(
                            Conversion(network.get_node(secondName), network.get_node(secondName.upper()), lambda x: x))
                        network.add_edge(
                            Conversion(network.get_node(secondName.upper()), network.get_node(secondName), lambda x: x))
                        dictionary[unit_kind.split('_')[0]].append(secondName.upper())
            else:
                for unit_name in list(dictionary[unit_kind]):
                    # print('U  2: ' + unit_name,unit_kind.split('_')[0])
                    network.add_node(UNode(unit_name))
                    network.add_node(UNode(unit_name.upper()))
                    network.add_edge(
                        Conversion(network.get_node(unit_name), network.get_node(unit_name.upper()), lambda x: x))
                    network.add_edge(
                        Conversion(network.get_node(unit_name.upper()), network.get_node(unit_name), lambda x: x))
                    dictionary[unit_kind.split('_')[0]].append(unit_name.upper())
        if '_INVERSE' in unit_kind:
            pass

    # Percentage & fraction :
    network.add_edge(Conversion(network.get_node('fraction'), network.get_node('percentage'), lambda x: x * 100))
    network.add_edge(Conversion(network.get_node('percentage'), network.get_node('fraction'), lambda p: p / 100))

    # Time conversions
    # network.addEdge(Conversion(network.getNode('second'), network.getNode('millisecond'), lambda t: t*1000))
    network.add_edge(Conversion(network.get_node('minute'), network.get_node('second'), lambda t: t * 60))
    network.add_edge(Conversion(network.get_node('hour'), network.get_node('minute'), lambda t: t * 60))
    network.add_edge(Conversion(network.get_node('day'), network.get_node('hour'), lambda t: t * 24))
    network.add_edge(Conversion(network.get_node('day'), network.get_node('month'), lambda t: t / 365.25 * 12))
    network.add_edge(Conversion(network.get_node('week'), network.get_node('day'), lambda t: t * 7))
    network.add_edge(Conversion(network.get_node('year'), network.get_node('month'), lambda t: t * 12))
    network.add_edge(Conversion(network.get_node('year'), network.get_node('day'), lambda t: t * 36525 / 100))
    network.add_edge(Conversion(network.get_node('lustrum'), network.get_node('year'), lambda t: t * 5))
    network.add_edge(Conversion(network.get_node('decade'), network.get_node('year'), lambda t: t * 10))
    network.add_edge(Conversion(network.get_node('century'), network.get_node('year'), lambda t: t * 100))

    # Temperature conversions
    network.add_edge(Conversion(network.get_node('Celsius'), network.get_node('Kelvin'), lambda t: t + 273.15))
    network.add_edge(Conversion(network.get_node('Kelvin'), network.get_node('Celsius'), lambda t: t - 273.15))
    network.add_edge(Conversion(network.get_node('Celsius'), network.get_node('Fahrenheit'), lambda t: t * 9 / 5 + 32))
    network.add_edge(
        Conversion(network.get_node('Fahrenheit'), network.get_node('Celsius'), lambda t: (t - 32) * 5 / 9))
    network.add_edge(Conversion(network.get_node('Fahrenheit'), network.get_node('Rankine'), lambda t: t + 459.67))
    network.add_edge(Conversion(network.get_node('Rankine'), network.get_node('Fahrenheit'), lambda t: t - 459.67))
    network.add_edge(Conversion(network.get_node('Rankine'), network.get_node('Kelvin'), lambda t: t * 5 / 9))
    network.add_edge(Conversion(network.get_node('Kelvin'), network.get_node('Rankine'), lambda t: t * 9 / 5))

    # Length conversions
    network.add_edge(Conversion(network.get_node('yard'), network.get_node('meter'), lambda d: d * 9144 / 10000))
    # network.addEdge(Conversion(network.getNode('foot'), network.getNode('meter'), lambda d: d*0.3048))
    network.add_edge(Conversion(network.get_node('inch'), network.get_node('thou'), lambda d: d * 1000))
    network.add_edge(Conversion(network.get_node('foot'), network.get_node('inch'), lambda d: d * 12))
    network.add_edge(Conversion(network.get_node('yard'), network.get_node('foot'), lambda d: d * 3))
    network.add_edge(Conversion(network.get_node('chain'), network.get_node('yard'), lambda d: d * 22))
    network.add_edge(Conversion(network.get_node('furlong'), network.get_node('chain'), lambda d: d * 10))
    network.add_edge(Conversion(network.get_node('mile'), network.get_node('furlong'), lambda d: d * 8))
    network.add_edge(Conversion(network.get_node('league'), network.get_node('mile'), lambda d: d * 3))
    network.add_edge(
        Conversion(network.get_node('nautical league'), network.get_node('nautical mile'), lambda d: d * 3))
    network.add_edge(Conversion(network.get_node('nautical mile'), network.get_node('meter'), lambda d: d * 1852))
    network.add_edge(Conversion(network.get_node('rod'), network.get_node('yard'), lambda d: d * 55 / 10))

    # Area conversions
    network.add_edge(Conversion(network.get_node('square mile'), network.get_node('acre'), lambda d: d * 640))
    network.add_edge(Conversion(network.get_node('acre'), network.get_node('square yard'), lambda d: d * 4840))
    network.add_edge(
        Conversion(network.get_node('square rod'), network.get_node('square yard'), lambda d: d * 3025 / 100))
    network.add_edge(Conversion(network.get_node('square yard'), network.get_node('square foot'), lambda d: d * 9))
    network.add_edge(Conversion(network.get_node('square foot'), network.get_node('square inch'), lambda d: d * 144))
    network.add_edge(Conversion(network.get_node('square foot'), network.get_node('square meter'),
                                lambda d: d * (3048 ** 2) / (10000 ** 2)))
    network.add_edge(Conversion(network.get_node('Darcy'), network.get_node('mD'), lambda d: d * 1000))
    network.add_edge(Conversion(network.get_node('Darcy'), network.get_node('µm2'), lambda d: d * 0.9869233))
    # network.addEdge(Conversion(network.getNode('m*m'), network.getNode('m'), lambda d: d**0.5))
    # network.addEdge(Conversion(network.getNode('m'), network.getNode('m*m'), lambda d: d**2))
    # network.addEdge(Conversion(network.getNode('rd*rd'), network.getNode('rd'), lambda d: d**0.5))
    # network.addEdge(Conversion(network.getNode('rd'), network.getNode('rd*rd'), lambda d: d**2))
    # network.addEdge(Conversion(network.getNode('yd*yd'), network.getNode('yd'), lambda d: d**0.5))
    # network.addEdge(Conversion(network.getNode('yd'), network.getNode('yd*yd'), lambda d: d**2))
    # network.addEdge(Conversion(network.getNode('ft*ft'), network.getNode('ft'), lambda d: d**0.5))
    # network.addEdge(Conversion(network.getNode('ft'), network.getNode('ft*ft'), lambda d: d**2))
    # network.addEdge(Conversion(network.getNode('in*in'), network.getNode('in'), lambda d: d**0.5))
    # network.addEdge(Conversion(network.getNode('in'), network.getNode('in*in'), lambda d: d**2))

    # Volume conversions
    network.add_edge(Conversion(network.get_node('gill'), network.get_node('fluid ounce'), lambda v: v * 4))
    network.add_edge(Conversion(network.get_node('pint'), network.get_node('gill'), lambda v: v * 4))
    network.add_edge(Conversion(network.get_node('quart'), network.get_node('pint'), lambda v: v * 2))
    network.add_edge(Conversion(network.get_node('gallonUS'), network.get_node('fluid ounce'), lambda v: v * 128))
    network.add_edge(Conversion(network.get_node('gallonUS'), network.get_node('quart'), lambda v: v * 4))
    network.add_edge(Conversion(network.get_node('gallonUS'), network.get_node('cubic inch'), lambda v: v * 231))

    network.add_edge(Conversion(network.get_node('gallonUK'), network.get_node('quartUK'), lambda v: v * 4))
    network.add_edge(Conversion(network.get_node('gallonUK'), network.get_node('fluid ounce UK'), lambda v: v * 160))
    network.add_edge(Conversion(network.get_node('gallonUK'), network.get_node('litre'), lambda v: v * 4.54609))
    network.add_edge(Conversion(network.get_node('gillUK'), network.get_node('fluid ounce UK'), lambda v: v * 4))
    network.add_edge(Conversion(network.get_node('pintUK'), network.get_node('gillUK'), lambda v: v * 4))
    network.add_edge(Conversion(network.get_node('quartUK'), network.get_node('pintUK'), lambda v: v * 2))

    network.add_edge(Conversion(network.get_node('gallonUK'), network.get_node('liter'), lambda v: v * 4.54609))
    network.add_edge(Conversion(network.get_node('cubic foot'), network.get_node('cubic meter'),
                                lambda v: v * (3048 ** 3) / (10000 ** 3)))
    network.add_edge(Conversion(network.get_node('standard cubic foot'), network.get_node('standard cubic meter'),
                                lambda v: v * (3048 ** 3) / (10000 ** 3)))
    network.add_edge(Conversion(network.get_node('standard barrel'), network.get_node('USgal'), lambda v: v * 42))
    network.add_edge(
        Conversion(network.get_node('standard cubic meter'), network.get_node('standard barrel'),
                   lambda v: v * 6.289814))
    network.add_edge(
        Conversion(network.get_node('standard barrel'), network.get_node('standard cubic foot'),
                   lambda v: v * 5.614584))
    network.add_edge(Conversion(network.get_node('reservoir cubic meter'), network.get_node('reservoir barrel'),
                                lambda v: v * 6.289814))
    network.add_edge(Conversion(network.get_node('reservoir cubic meter'), network.get_node('standard cubic meter'),
                                lambda v: v / network.get_fvf()))
    # network.addEdge(Conversion(network.getNode('standard cubic meter'), network.getNode('standard cubic foot'), lambda v: v/5.614584))

    # Pressure conversions
    network.add_edge(Conversion(network.get_node('psi gauge'), network.get_node('absolute psi'), lambda p: p + 14.6959))
    network.add_edge(Conversion(network.get_node('absolute psi'), network.get_node('psi gauge'), lambda p: p - 14.6959))
    network.add_edge(Conversion(network.get_node('bar gauge'), network.get_node('absolute bar'), lambda p: p + 1.01325))
    network.add_edge(Conversion(network.get_node('absolute bar'), network.get_node('bar gauge'), lambda p: p - 1.01325))

    network.add_edge(
        Conversion(network.get_node('absolute bar'), network.get_node('absolute psi'), lambda p: p * 14.50377377322))
    network.add_edge(
        Conversion(network.get_node('bar gauge'), network.get_node('psi gauge'), lambda p: p * 14.50377377322))
    network.add_edge(Conversion(network.get_node('bar'), network.get_node('psi'), lambda p: p * 14.50377377322))
    network.add_edge(Conversion(network.get_node('psi'), network.get_node('bar'), lambda p: p / 14.50377377322))
    network.add_edge(Conversion(network.get_node('absolute bar'), network.get_node('Pascal'), lambda p: p * 100000))
    # network.addEdge(Conversion(network.getNode('atmosphere'), network.getNode('absolute bar'), lambda p: p*1.01325))
    network.add_edge(Conversion(network.get_node('atmosphere'), network.get_node('Pascal'), lambda p: p * 101325))
    network.add_edge(Conversion(network.get_node('atmosphere'), network.get_node('Torr'), lambda p: p * 760))
    network.add_edge(Conversion(network.get_node('absolute bar'), network.get_node('bar'), lambda p: p))
    network.add_edge(Conversion(network.get_node('absolute psi'), network.get_node('psi'), lambda p: p))
    network.add_edge(Conversion(network.get_node('bar gauge'), network.get_node('bar'), lambda p: p))
    network.add_edge(Conversion(network.get_node('psi gauge'), network.get_node('psi'), lambda p: p))

    # mass Conversion
    network.add_edge(Conversion(network.get_node('grain'), network.get_node('milligrams'), lambda w: w * 64.7989))
    network.add_edge(Conversion(network.get_node('pennyweight'), network.get_node('grain'), lambda w: w * 24))
    network.add_edge(Conversion(network.get_node('dram'), network.get_node('pound'), lambda w: w / 256))
    network.add_edge(Conversion(network.get_node('stone'), network.get_node('pound'), lambda w: w * 14))
    network.add_edge(Conversion(network.get_node('quarter'), network.get_node('stone'), lambda w: w * 2))
    network.add_edge(Conversion(network.get_node('ounce'), network.get_node('dram'), lambda w: w * 16))
    network.add_edge(Conversion(network.get_node('pound'), network.get_node('ounce'), lambda w: w * 16))
    network.add_edge(Conversion(network.get_node('long hundredweight'), network.get_node('quarter'), lambda w: w * 4))
    network.add_edge(Conversion(network.get_node('short hundredweight'), network.get_node('pound'), lambda w: w * 100))
    network.add_edge(
        Conversion(network.get_node('short ton'), network.get_node('short hundredweight'), lambda w: w * 20))
    network.add_edge(Conversion(network.get_node('long ton'), network.get_node('long hundredweight'), lambda w: w * 20))

    network.add_edge(Conversion(network.get_node('metric ton'), network.get_node('kilogram'), lambda w: w * 1000))
    network.add_edge(Conversion(network.get_node('kilogram'), network.get_node('gram'), lambda w: w * 1000))
    # network.addEdge(Conversion(network.getNode('pound'), network.getNode('gram'), lambda w: w*453.59237))
    network.add_edge(
        Conversion(network.get_node('pound'), network.get_node('kilogram'), lambda w: w * 45359237 / 100000000))

    # force Conversion
    # network.addEdge(Conversion(network.getNode('kilogram'), network.getNode('kilogram force'), lambda f: f* converter(StandardEarthGravity,'m/s2','cm/s2',False)))
    network.add_edge(Conversion(network.get_node('kilogram mass'), network.get_node('kilogram force'),
                                lambda f: f * StandardEarthGravity))
    network.add_edge(Conversion(network.get_node('kilogram force'), network.get_node('kilogram mass'),
                                lambda f: f / StandardEarthGravity))
    network.add_edge(Conversion(network.get_node('Dyne'), network.get_node('Newton'), lambda f: f * 1E-5))
    network.add_edge(Conversion(network.get_node('Newton'), network.get_node('Dyne'), lambda f: f * 1E5))

    # Energy Conversion
    network.add_edge(Conversion(network.get_node('Joule'), network.get_node('gram calorie'), lambda e: e / 4.184))
    network.add_edge(Conversion(network.get_node('Kilojoule'), network.get_node('Joule'), lambda e: e * 1000))
    network.add_edge(Conversion(network.get_node('Kilojoule'), network.get_node('kilowatt hour'), lambda e: e / 3600))
    network.add_edge(
        Conversion(network.get_node('Kilojoule'), network.get_node('British thermal unit'), lambda e: e / 1.055))

    # Power Conversion
    network.add_edge(Conversion(network.get_node('Horsepower'), network.get_node('Watt'), lambda e: e * 745.699872))

    # Density Conversion
    network.add_edge(Conversion(network.get_node('API'), network.get_node('SgO'), lambda d: 141.5 / (131.5 + d)))
    network.add_edge(Conversion(network.get_node('SgO'), network.get_node('API'), lambda d: 141.5 / d - 131.5))
    network.add_edge(Conversion(network.get_node('API'), network.get_node('g/cc'), lambda d: 141.5 / (131.5 + d)))
    network.add_edge(Conversion(network.get_node('g/cc'), network.get_node('API'), lambda d: 141.5 / d - 131.5))
    network.add_edge(Conversion(network.get_node('SgO'), network.get_node('g/cc'), lambda d: d))
    network.add_edge(Conversion(network.get_node('SgW'), network.get_node('g/cc'), lambda d: d))
    network.add_edge(Conversion(network.get_node('SgG'), network.get_node('g/cc'), lambda d: d * StandardAirDensity))
    network.add_edge(Conversion(network.get_node('psia/ft'), network.get_node('lb/ft3'), lambda d: d * 144))
    network.add_edge(
        Conversion(network.get_node('g/cm3'), network.get_node('lb/ft3'), lambda d: d * 62.427960576144606))
    network.add_edge(Conversion(network.get_node('lb/ft3'), network.get_node('lb/stb'), lambda d: d * 5.614584))

    # viscosity conversions
    network.add_edge(Conversion(network.get_node('Pa*s'), network.get_node('Poise'), lambda v: v * 10))

    # data conversions
    network.add_edge(Conversion(network.get_node('byte'), network.get_node('bit'), lambda d: d * 8))
    network.add_edge(Conversion(network.get_node('byte'), network.get_node('B'), lambda d: d))
    network.add_edge(Conversion(network.get_node('bit'), network.get_node('b'), lambda d: d))

    for unit_kind in list(dictionary.keys()):
        if '_REVERSE' in unit_kind:
            if type(dictionary[unit_kind]) is dict:
                nameList = list(dictionary[unit_kind].keys())
            else:
                nameList = list(dictionary[unit_kind])
            # print(nameList)
            for unit_name in nameList:
                # print('R  2: ' + unit_name)
                for otherName in network.children_of(network.get_node(unit_name)):
                    # print('R   3: '+unit_name,otherName.getName())
                    if network.get_node(unit_name) != otherName:
                        network.add_edge(Conversion(otherName, network.get_node(unit_name),
                                                    network.edges[network.get_node(unit_name)][1][
                                                        network.edges[network.get_node(unit_name)][0].index(otherName)],
                                                    True))

    for unit_kind in list(dictionary.keys()):
        if '_FROMvolume' in unit_kind and unit_kind.split('_')[0] in SI_order[2]:
            # if '_SI' in unit_kind and unit_kind.split('_')[0]  in SI_order[2] :
            for unit_name in list(dictionary[unit_kind]):
                # print('S  2: ' + unit_name)
                network.add_node(UNode(unit_name))
                dictionary[unit_kind.split('_')[0]].append(unit_name)
                for otherName in network.children_of(network.get_node(unit_name.split('/')[0])):
                    if network.get_node(unit_name.split('/')[0]) != otherName:
                        print('R   3: ' + unit_name, otherName.get_name())
                        otherRate = otherName.get_name() + '/' + unit_name.split('/')[1]
                        network.add_node(UNode(otherRate))
                        network.add_edge(Conversion(network.get_node(unit_name), otherRate,
                                                    network.edges[network.get_node(unit_name.split('/')[1])][1][
                                                        network.edges[network.get_node(unit_name.split('/')[1])][
                                                            0].index(
                                                            otherName)]))
                        network.add_edge(Conversion(otherRate, network.get_node(unit_name),
                                                    network.edges[network.get_node(unit_name.split('/')[1])][1][
                                                        network.edges[network.get_node(unit_name.split('/')[1])][
                                                            0].index(
                                                            otherName)], True))

    to_remove = []
    for unit_kind in dictionary:
        if '_' in unit_kind:
            to_remove.append(unit_kind)
        else:  # if '_' not in unit_kind :
            dictionary[unit_kind] = tuple(dictionary[unit_kind])
    dictionary['UserUnits'] = []

    for unit_kind in to_remove:
        dictionary.pop(unit_kind)

    return network


def _create_Rates() -> None:
    # volumes / Time
    rates = list(dictionary['Rate']) if 'Rate' in dictionary else []
    rates += [volume + '/' + time for volume in dictionary['Volume'] for time in dictionary['Time']]
    rates += [weight + '/' + time for weight in dictionary['Weight'] for time in dictionary['Time']]
    rates += [data + '/' + time for data in dictionary['dataBYTE'] for time in dictionary['Time']]
    rates += [data + '/' + time for data in dictionary['dataBIT'] for time in dictionary['Time']]
    dictionary['Rate'] = tuple(set(rates))


def _create_VolumeRatio() -> None:
    # Volume / Volume
    ratio = list(dictionary['VolumeRatio']) if 'VolumeRatio' in dictionary else []
    ratio += [numerator + '/' + denominator for numerator in dictionary['Volume'] for denominator in
              dictionary['Volume']]
    dictionary['VolumeRatio'] = tuple(set(ratio))


def _create_Density() -> None:
    # mass / Volume
    density = list(dictionary['Density']) if 'Density' in dictionary else []
    density += [mass + '/' + volume for mass in dictionary['Mass'] for volume in dictionary['Volume']]
    dictionary['Density'] = tuple(set(density))


def _create_Speed() -> None:
    # Length / Time
    speed = list(dictionary['Speed']) if 'Speed' in dictionary else []
    speed += [length + '/' + time for length in dictionary['Length'] for time in dictionary['Time']]
    dictionary['Speed'] = tuple(set(speed))


def _create_Power() -> None:
    # Length / Time
    power = list(dictionary['Power']) if 'Power' in dictionary else []
    power += [energy + '/' + time for energy in dictionary['Energy'] for time in dictionary['Time']]
    dictionary['Power'] = tuple(set(power))


def _create_ProductivityIndex() -> None:
    # Volume / Time / Pressure
    productivityIndex = list(dictionary['ProductivityIndex']) if 'ProductivityIndex' in dictionary else []
    productivityIndex += [volume + '/' + time + '/' + pressure
                          for volume in dictionary['Volume']
                          for time in dictionary['Time']
                          for pressure in dictionary['Pressure']]
    dictionary['ProductivityIndex'] = tuple(set(productivityIndex))


def _create_PressureGradient() -> None:
    # Pressure / Length
    pressureGradient = list(dictionary['PressureGradient']) if 'PressureGradient' in dictionary else []
    pressureGradient += [pressure + '/' + length for pressure in dictionary['Pressure'] for length in
                         dictionary['Length']]
    dictionary['PressureGradient'] = tuple(set(pressureGradient))


def _create_TemperatureGradient() -> None:
    # Pressure / Length
    temperatureGradient = list(dictionary['TemperatureGradient']) if 'TemperatureGradient' in dictionary else []
    temperatureGradient += [temperature + '/' + length for temperature in dictionary['Temperature'] for length in
                            dictionary['Length']]
    dictionary['TemperatureGradient'] = tuple(set(temperatureGradient))


def _create_Acceleration() -> None:
    # Length / Time / Time
    acceleration = list(dictionary['Acceleration']) if 'Acceleration' in dictionary else []
    acceleration += [(length + '/' + time1 + '2') if time1 == time2 else (length + '/' + time1 + '/' + time2)
                     for length in dictionary['Length']
                     for time1 in dictionary['Time']
                     for time2 in dictionary['Time']]
    dictionary['Acceleration'] = tuple(set(acceleration))


def rebuild_units():
    from .dictionaries import _load_dictionary
    dictionary, temperatureRatioConversions = _load_dictionary()
    units_network = _load_network()
    unyts_parameters_.reload_ = True
    unyts_parameters_.save_params()
    return units_network, dictionary, temperatureRatioConversions


def network2frame() -> DataFrame:
    frame = DataFrame(data={}, columns=['source', 'target', 'lambda'])
    i = 0
    for node in unitsNetwork.edges:
        for children in unitsNetwork.children_of(node):
            frame.loc[i, ['source', 'target', 'lambda']] = [node.get_name(), children.get_name(),
                                                            unitsNetwork.conversion(node, children)]
            i += 1
    return frame.drop_duplicates(['source', 'target'])


# load the network into an instance of the graph database
if not unyts_parameters_.reload_ and \
        isfile(dir_path + 'units/UnitsNetwork.cache') and \
        isfile(dir_path + 'units/UnitsDictionary.cache') and \
        isfile(dir_path + 'units/TemperatureRatioConversions.cache'):
    try:
        with open(dir_path + 'units/UnitsNetwork.cache', 'rb') as f:
            unitsNetwork = load(f)
        print('units network loaded from cache...')
        unyts_parameters_.reload_ = False
        unyts_parameters_.save_params()
    except:
        unitsNetwork, dictionary, temperatureRatioConversions = rebuild_units()
else:
    try:
        unitsNetwork = _load_network()
        # load the dictionary with ratio unis
        _create_Rates()
        _create_VolumeRatio()
        _create_Density()
        _create_Speed()
        _create_ProductivityIndex()
        _create_PressureGradient()
        _create_TemperatureGradient()
        _create_Power()
        unyts_parameters_.reload_ = False
        unyts_parameters_.save_params()
    except:
        unitsNetwork, dictionary, temperatureRatioConversions = rebuild_units()
    if unyts_parameters_.cache_:
        print('saving units network and dictionary to cache...')
        with open(dir_path + 'units/UnitsNetwork.cache', 'wb') as f:
            dump(unitsNetwork, f)
        with open(dir_path + 'units/UnitsDictionary.cache', 'w') as f:
            jdump(dictionary, f)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 14:34:59 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['Unit']

from unyts.errors import WrongUnitsError, WrongValueError, NoConversionFoundError
from unyts.operations import unit_product, unit_division, unit_base_power
from unyts.converter import convert, convertible
import numpy as np
from numbers import Number
from numpy import ndarray
from typing import Union

try:
    from pandas import Series, DataFrame
    array_like = tuple([ndarray, Series, DataFrame])
    numeric = Union[int, float, complex, ndarray, Series, DataFrame]
except ModuleNotFoundError:
    array_like = tuple([ndarray])
    numeric = Union[int, float, complex, ndarray]
number = Union[int, float, complex]


class Unit(object):
    """
    A class to cope with values associated to units, its arithmetic and logic
    operations and conversions.
    """

    def __init__(self, value=None, unit=None):
        self.unit = None
        self.value = None
        self.name = None
        self.kind = None
        if value is not None and unit is not None:
            from .units.define import units
            u = units(value, unit)
            self.unit = u.unit
            self.value = u.value
            self.name = u.name
            self.kind = u.kind

    def __call__(self) -> numeric:
        return self.value

    def __repr__(self) -> str:
        return str(self.value) + '_' + str(self.unit)

    def __str__(self) -> str:
        return str(self.value) + '_' + str(self.unit)

    def convert(self, new_unit):
        if type(new_unit) is not str:
            try:
                new_unit = new_unit.unit
            except AttributeError:
                raise WrongUnitsError("'" + str(new_unit) + "' for '" + str(self.name) + "'")
        return self.kind(convert(self.value, self.unit, new_unit), new_unit)

    def to(self, new_unit):
        return self.convert(new_unit)

    def __neg__(self):
        return self.kind(self.value * -1, self.unit)

    def __bool__(self):
        from .units.custom import UserUnits
        from .units.unitless import Dimensionless, Percentage
        if self.kind in (Dimensionless, Percentage, UserUnits):
            return False
        return True

    def __abs__(self):
        return self.kind(abs(self.value), self.unit)

    def __add__(self, other):
        from .units.unitless import Dimensionless, Percentage
        if '.units.' in str(type(other)):
            if other.kind is self.kind:
                if self.unit == other.unit:
                    return self.kind(self.value + other.value, self.unit)
                elif convertible(other.unit, self.unit):
                    return self.kind(self.value + convert(other.value, other.unit, self.unit), self.unit)
                elif convertible(self.unit, other.unit):
                    return self.kind(other.value + convert(self.value, self.unit, other.unit), other.unit)
                else:
                    raise NoConversionFoundError("from '" + self.unit + "' to '" + other.unit + "'")
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return self.kind(self.value + other.value, other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value + other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value + other.value, self.unit)
            else:  # add different units
                return self, other
        elif type(other) is tuple:
            result, flag = [], False
            for each in other:
                if type(each) is type(self):
                    result += [self + each]
                    flag = True
                else:
                    result += each
            return tuple(result) if flag else other + (self,)
        elif self.kind is Percentage:
            return self.kind((self.value + other) * 100, self.unit)
        else:
            return self.kind(self.value + other, self.unit)

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if '.units.' in str(type(other)):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value * other.value, other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value * other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value * other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value * other.value, unit_product(self.unit, other.unit))
            elif convertible(other.unit, self.unit):
                return units(self.value * convert(other.value, other.unit, self.unit),
                             unit_product(self.unit, other.unit))
            elif convertible(self.unit, other.unit):
                return units(other.value * convert(self.value, self.unit, other.unit),
                             unit_product(other.unit, self.unit))
            elif convertible(unit_base_power(self.unit)[0], unit_base_power(other.unit)[0]):
                factor = convert(1, unit_base_power(other.unit)[0], unit_base_power(self.unit)[0])
                return units(self.value * other.value * factor, unit_product(self.unit, other.unit))
            else:
                return units(self.value * other.value, unit_product(self.unit, other.unit))
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self * each]
            return tuple(result)
        elif self.kind is Percentage:
            return self.kind(self.value * other * 100, self.unit)
        else:
            return self.kind(self.value * other, self.unit)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __pow__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if '.units.' in str(type(other)):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value ** other.value, other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value ** other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value ** other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value ** other.value, self.unit + '^' + other.unit)
            elif convertible(other.unit, self.unit):
                return units(self.value ** convert(other.value, other.unit, self.unit), self.unit + '^' + self.unit)
            elif convertible(unit_base_power(self.unit)[0], unit_base_power(other.unit)[0]):
                factor = convert(1, unit_base_power(other.unit)[0], unit_base_power(self.unit)[0])
                return units(self.value ** (other.value * factor), self.unit + '^' + other.unit)
            else:
                return units(self.value ** other.value, self.unit + '^' + other.unit)
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self ** each]
            return tuple(result)
        else:
            if self.kind is Percentage:
                return self.kind((self.value ** other) * 100, self.unit)
            else:
                pow_units = unit_base_power(self.unit)[1] * other
                if pow_units == 0:
                    pow_units = 'Dimensionless'
                elif pow_units == 1:
                    pow_units = unit_base_power(self.unit)[0]
                else:
                    pow_units = unit_base_power(self.unit)[0] + str(pow_units)
                return units(self.value ** other, pow_units)

    def __rpow__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if self.kind is Percentage:
            return units((other ** self.value), 'Dimensionless')
        elif self.kind is Dimensionless:
            return units(other ** self.value, self.unit)
        else:
            raise TypeError("unsupported operand type(s) for ** or pow(): '" + str(type(other)) + "' and '" + self.name)

    def __sub__(self, other):
        # return self.__add__(other * -1)
        from .units.unitless import Dimensionless, Percentage
        if '.units.' in str(type(other)):
            if other.kind is self.kind:
                if self.unit == other.unit:
                    return self.kind(self.value - other.value, self.unit)
                elif convertible(other.unit, self.unit):
                    return self.kind(self.value - convert(other.value, other.unit, self.unit), self.unit)
                elif convertible(self.unit, other.unit):
                    return self.kind(other.value - convert(self.value, self.unit, other.unit), other.unit)
                else:
                    raise NoConversionFoundError("from '" + self.unit + "' to '" + other.unit + "'")
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return self.kind(self.value - other.value, other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value - other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value - other.value, self.unit)
            else:  # add different units
                return self, other
        elif type(other) is tuple:
            result, flag = [], False
            for each in other:
                if type(each) is type(self):
                    result += [self - each]
                    flag = True
                else:
                    result += each
            return tuple(result) if flag else other + (self,)
        elif self.kind is Percentage:
            return self.kind((self.value - other) * 100, self.unit)
        else:
            return self.kind(self.value - other, self.unit)

    def __rsub__(self, other):
        return self.__neg__() + other

    def __truediv__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if '.units.' in str(type(other)):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value / other.value, '1/' + other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value / other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value / other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value / other.value, unit_division(self.unit, other.unit))
            elif convertible(other.unit, self.unit):
                return units(self.value / convert(other.value, other.unit, self.unit),
                             unit_division(self.unit, other.unit))
            elif convertible(unit_base_power(self.unit)[0], unit_base_power(other.unit)[0]):
                factor = convert(1, unit_base_power(other.unit)[0], unit_base_power(self.unit)[0])
                return units(self.value / (other.value * factor), unit_division(self.unit, other.unit))
            else:
                return units(self.value / other.value, unit_division(self.unit, other.unit))
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self / each]
            return tuple(result)
        elif self.kind is Percentage:
            return self.kind(self.value / other * 100, self.unit)
        else:
            return units(self.value / other, self.unit)

    def __rtruediv__(self, other):
        from .units.define import units
        from .units.unitless import Dimensionless, Percentage
        if self.kind is Percentage:
            return units(other / self.value * 100, self.unit)
        elif self.kind is Dimensionless:
            return units(other / self.value, self.unit)
        else:
            return units(other / self.value, self.unit + '-1')

    def __floordiv__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if '.units.' in str(type(other)):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value // other.value, '1/' + other.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value // other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value // other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value // other.value, unit_division(self.unit, other.unit))
            elif convertible(other.unit, self.unit):
                return units(self.value // convert(other.value, other.unit, self.unit),
                             unit_division(self.unit, other.unit))
            elif convertible(unit_base_power(self.unit)[0], unit_base_power(other.unit)[0]):
                factor = convert(1, unit_base_power(other.unit)[0], unit_base_power(self.unit)[0])
                return units(self.value // (other.value * factor), unit_division(self.unit, other.unit))
            else:
                return units(self.value // other.value, unit_division(self.unit, other.unit))
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self // each]
            return tuple(result)
        elif self.kind is Percentage:
            return self.kind(self.value // other * 100, self.unit)
        else:
            return self.kind(self.value // other, self.unit)

    def __rfloordiv__(self, other):
        from .units.define import units
        from .units.unitless import Dimensionless, Percentage
        if self.kind is Percentage:
            return units(other // self.value * 100, self.unit)
        elif self.kind is Dimensionless:
            return units(other // self.value, self.unit)
        else:
            return units(other // self.value, self.unit + '-1')

    def __mod__(self, other):
        from .units.unitless import Dimensionless, Percentage
        from .units.define import units
        if '.units.' in str(type(other)):
            if self.kind in (Dimensionless, Percentage) and other.kind not in (Dimensionless, Percentage):
                return other.kind(self.value % other.value, self.unit)
            elif self.kind is Percentage and other.kind in (Dimensionless, Percentage):
                return self.kind((self.value % other.value) * 100, self.unit)
            elif other.kind in (Dimensionless, Percentage):
                return self.kind(self.value % other.value, self.unit)
            elif self.unit == other.unit:
                return units(self.value % other.value, self.unit)
            elif convertible(other.unit, self.unit):
                return units(self.value % convert(other.value, other.unit, self.unit), self.unit)
            elif convertible(unit_base_power(self.unit)[0], unit_base_power(other.unit)[0]):
                factor = convert(1, unit_base_power(other.unit)[0], unit_base_power(self.unit)[0])
                return units(self.value % (other.value * factor), self.unit)
            else:
                return units(self.value % other.value, self.unit)
        elif type(other) is tuple:
            result = []
            for each in other:
                result += [self % each]
            return tuple(result)
        elif self.kind is Percentage:
            return self.kind(self.value % other * 100, self.unit)
        else:
            return self.kind(self.value % other, self.unit)

    def __lt__(self, other) -> bool:
        if type(self) == type(other):
            return self.value < other.convert(self.unit).value
        else:
            msg = "'<' not supported between instances of '" + \
                  (str(type(self))[
                   str(type(self)).index("'") + 1:len(str(type(self))) - str(type(self))[::-1].index("'") - 1]).replace(
                      '__main__.', '') + \
                  "' and '" + \
                  (str(type(other))[
                   str(type(other)).index("'") + 1:len(str(type(other))) - str(type(other))[::-1].index(
                       "'") - 1]).replace('__main__.', '') + "'"
            raise TypeError(msg)

    def __le__(self, other) -> bool:
        if type(self) == type(other):
            return self.value <= other.convert(self.unit).value
        else:
            msg = "'<=' not supported between instances of '" + \
                  (str(type(self))[
                   str(type(self)).index("'") + 1:len(str(type(self))) - str(type(self))[::-1].index("'") - 1]).replace(
                      '__main__.', '') + \
                  "' and '" + \
                  (str(type(other))[
                   str(type(other)).index("'") + 1:len(str(type(other))) - str(type(other))[::-1].index(
                       "'") - 1]).replace('__main__.', '') + "'"
            raise TypeError(msg)

    def __eq__(self, other) -> bool:
        if type(self) == type(other):
            return self.value == other.convert(self.unit).value
        else:
            msg = "'==' not supported between instances of '" + \
                  (str(type(self))[
                   str(type(self)).index("'") + 1:len(str(type(self))) - str(type(self))[::-1].index("'") - 1]).replace(
                      '__main__.', '') + \
                  "' and '" + \
                  (str(type(other))[
                   str(type(other)).index("'") + 1:len(str(type(other))) - str(type(other))[::-1].index(
                       "'") - 1]).replace('__main__.', '') + "'"
            raise TypeError(msg)

    def __ne__(self, other):
        if type(self) == type(other):
            return self.value != other.convert(self.unit).value
        else:
            msg = "'!=' not supported between instances of '" + \
                  (str(type(self))[
                   str(type(self)).index("'") + 1:len(str(type(self))) - str(type(self))[::-1].index("'") - 1]).replace(
                      '__main__.', '') + \
                  "' and '" + \
                  (str(type(other))[
                   str(type(other)).index("'") + 1:len(str(type(other))) - str(type(other))[::-1].index(
                       "'") - 1]).replace('__main__.', '') + "'"
            raise TypeError(msg)

    def __ge__(self, other):
        if type(self) == type(other):
            return self.value >= other.convert(self.unit).value
        else:
            msg = "'>=' not supported between instances of '" + \
                  (str(type(self))[
                   str(type(self)).index("'") + 1:len(str(type(self))) - str(type(self))[::-1].index("'") - 1]).replace(
                      '__main__.', '') + \
                  "' and '" + \
                  (str(type(other))[
                   str(type(other)).index("'") + 1:len(str(type(other))) - str(type(other))[::-1].index(
                       "'") - 1]).replace('__main__.', '') + "'"
            raise TypeError(msg)

    def __gt__(self, other):
        if type(self) == type(other):
            return self.value > other.convert(self.unit).value
        else:
            msg = "'>' not supported between instances of '" + \
                  (str(type(self))[
                   str(type(self)).index("'") + 1:len(str(type(self))) - str(type(self))[::-1].index("'") - 1]).replace(
                      '__main__.', '') + \
                  "' and '" + \
                  (str(type(other))[
                   str(type(other)).index("'") + 1:len(str(type(other))) - str(type(other))[::-1].index(
                       "'") - 1]).replace('__main__.', '') + "'"
            raise TypeError(msg)

    def __len__(self):
        try:
            return len(self.value)
        except TypeError:
            return 1

    def __getitem__(self, item):
        if type(item) is int:
            if item >= len(self):
                raise IndexError
        else:
            raise ValueError
        from .units.unitless import Percentage
        if self.kind is Percentage:
            return self.kind(self.value[item] * 100, self.unit)
        else:
            return self.kind(self.value[item], self.unit)

    def __iter__(self):
        if type(self.value) in (int, float):
            return np.array((self.value,)).__iter__()
        else:
            return self.value.__iter__()

    def get_unit(self):
        return self.unit

    def get_value(self):
        return self.value

    def check_value(self, value):
        if type(value) in (list, tuple):
            try:
                return np.array(value)
            except TypeError:
                raise WrongValueError(str(value))
        elif isinstance(value, Number):
            return value
        elif isinstance(value, array_like):
            return value
        else:
            raise WrongValueError(str(value))

    def check_unit(self, units):
        if type(units) is not str:
            try:
                units = units.unit
            except AttributeError:
                raise WrongUnitsError("'" + str(units) + "' for '" + str(self.name) + "'")
        if units in self.kind.classUnits:
            return units
        else:
            raise WrongUnitsError("'" + str(units) + "' for '" + str(self.name) + "'")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 24 15:57:27 2020

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['_split_ratio', '_split_product', '_split_unit', '_reduce_parentheses']


from unyts.helpers.multi_split import multi_split


def _split_ratio(unit: str) -> tuple[str]:
    return tuple(map(str.strip, unit.split('/')))


def _split_product(unit: str) -> tuple[str]:
    return tuple(map(str.strip, unit.split('*')))


def _split_unit(unit: str) -> tuple[str]:
    return multi_split(unit,
                       sep=('*', '/',),
                       remove=None)


def _reduce_parentheses(unit: str) -> str:
    if '(' not in unit and ')' not in unit:
        return unit
    elif unit.count('(') > unit.count(')'):
        raise ValueError("closing parenthesis without opening parenthesis")
    elif unit.count('(') < unit.count(')'):
        raise ValueError("opening parenthesis without closing parenthesis")
    for o in ['^', '**', '+', '-']:
        if o in unit:
            print("not implemented to remove parenthesis when ' + o + ' is in the unit string")
            return unit

    unit_split = multi_split(unit,
                             sep=('+', '-', '*', '/', '**', '(', ')'),
                             remove=' ')
    result, inv, inp, ii, pa = [], False, False, 0, 0
    while ii < len(unit_split):
        if unit_split[ii] in '*/' and inp and inv:
            result.append('*' if unit_split[ii] == '/' else '/')
        elif unit_split[ii] not in '()':
            result.append(unit_split[ii])
        if unit_split[ii] == '/' and unit_split[ii + 1] == '(':
            inv = not inv
        elif unit_split[ii] == '(':
            inp, pa = True, pa + 1
        elif unit_split[ii] == ')':
            inp, pa = False, pa - 1
        ii += 1
    return ''.join(result)



# -*- coding: utf-8 -*-
"""
Created on Wed Aug  3 21:14:44 2022

@author: Martín Carlos Araya <martinaraya@gmail.com>
"""

__version__ = '0.4.9'
__release__ = 20221231
__all__ = ['caster', 'to_number']


from unyts.helpers.common_classes import number


def caster(string: str) -> number:
    """
    Helper function to try to cast the string to number

    Parameters
    ----------
    string : str

    Raises
    ------
    ValueError
        if not possible to cast to int, to float or to complex

    Returns
    -------
    value : int, float or complex
        the cast number

    """
    try:
        value = int(string)
    except ValueError:
        try:
            value = float(string)
        except ValueError:
            try:
                value = complex(string)
            except ValueError:
                raise ValueError('could not convert string to number: ' + string)
    return value


def to_number(string: str, decimal_sign='auto', thousand_separator='auto', parenthesis_means_negative=True) -> number:
    """
    Cast a string the appropriate numeric type.
    Valid numbers are:
        integers
        real numbers
        complex numbers

    By default, automatically converts comma ',' to period '.' and remove thousand separators

    Parameters
    ----------
    string : str
        the string to convert to number
    decimal_sign : str
        The string used as decimal symbol.
    thousand_separator : str
        The string used as thousands separator.
    parenthesis_means_negative : bool
        True to cast numbers between parenthesis as negative, i.e. (15) = -15

    Returns
    -------
    None.

    """

    def cleaner(string: str):
        """
        Helper function to clean the string before attempting to cast

        Parameters
        ----------
        string : str

        Returns
        -------
            string : the cleaned string

        """

        def validate(string: str, ds, ts):
            """
            Helper function to validate the string format

            Parameters
            ----------
            string : str
                the string to validate.
            ds : str
                the expected decimal sign.
            ts : str
                the expected thousand separator.

            Raises
            ------
            ValueError
                if string format is not valid.

            Returns
            -------
            True
                if string format is valid.

            """

            def check(string: str):
                """
                count the number of decimal signs and thousand separators
                """
                if len(string.split(ds)) > 2 :
                    raise ValueError('the string format is incorrect: ' + string)
                if ts is not None:
                    for part in string.split(ds)[0].split(ts)[1:]:
                        if len(part) != 3:
                            raise ValueError('the string format is incorrect: ' + string)
                if 'e' in string:
                    if 'j' in string:
                        if string.count('j') > 1 or string[-1] != 'j':
                            raise ValueError('the string format is incorrect: ' + string)
                        if not string.split('e')[-1][:-1].isdigit():
                            raise ValueError('the string format is incorrect: ' + string)

                    elif not string.split('e')[-1].isdigit():
                        raise ValueError('the string format is incorrect: ' + string)
                return True

            if 'j' in string:
                for s in ('+', '-'):
                    if s in string.strip('+-'):
                        return check(string.strip('+-').split(s)[0]) and check(string.strip('+-').split(s)[1])
            else:
                return check(string)

        string = string.strip().lower()
        new_string = string

        # check parenthesis means negative
        if parenthesis_means_negative:
            if '(' in new_string:
                if ')' not in new_string:
                    raise ValueError('the string format is incorrect: ' + string)
                elif 'j' in new_string:
                    pass
                else:
                    if new_string.count('(') == 1 and new_string.count(')') == 1:
                        if new_string[0] == '(' and new_string[-1] == ')':
                            if '-' not in new_string:
                                new_string = '-' + new_string.replace('(', '').replace(')', '')
                            else:
                                raise ValueError('the string format is incorrect: ' + string)
            elif ')' in new_string:
                raise ValueError('the string format is incorrect: ' + string)
            else:
                pass

        # check thousands separator
        if thousand_separator in ('auto', '.') and decimal_sign in ('auto', ','):
            if ',' in new_string and '.' in new_string:
                if validate(string,ds=',',ts='.'):
                    new_string = new_string.replace('.', '').replace(',', '.')
                return new_string

        if thousand_separator.strip() == 'auto':
            for ts in ("'","`","´"," ",None):
                if ts is not None and ts in string:
                    new_string = new_string.replace(ts, '')
                    break
        else:
            new_string = new_string.replace(thousand_separator, '')
            ts = thousand_separator

        # check decimal symbol
        if decimal_sign.strip() == 'auto':
            for ds in (",","p","."):
                if ds in new_string:
                    if new_string.count(ds) > 1 and 'j' not in new_string:
                        raise ValueError('the string format is incorrect: ' + string)
                    new_string = new_string.replace(ds, '.')
                    break
        else:
            if string.count(decimal_sign) > 1:
                raise ValueError('the string format is incorrect: ' + string)
            new_string = new_string.replace(decimal_sign, '.')
            ds = decimal_sign

        if validate(string, ds=ds, ts=ts):
            return new_string

    if type(string) is str:
        # first, clean the string if is type str
        string = cleaner(string)

        # then, cast the string
        value = caster(string)

        # convert to integer numbers in scientific notation without decimals digits
        if 'E' in string.upper() and '.' not in string:
            if type(value) is float:
                if int(value) == value:
                    value = int(value)

        return value

    elif type(string) in (int, float, complex):
        return string

    else:
        raise TypeError('string must be a str or a number')
